import os
import json
import atexit
from typing import Optional
from storages.sync_storages.base_nosql_storage import BaseNoSQLStorage


def create_dir(file_path):
    dirs, filename = os.path.split(file_path)
    os.makedirs(dirs, exist_ok=True)
    if not os.path.isfile(file_path):
        with open(file_path, 'w') as file:
            json.dump({}, file)


class JSONStorage(BaseNoSQLStorage):
    def __init__(self, file_path: Optional[str]):
        super().__init__()
        self.file_path = file_path if file_path is not None else "./.storages/data.json"
        create_dir(self.file_path)
        self.load()

        atexit.register(self.dump)

    def load(self):
        with open(self.file_path, 'r') as file:
            json_data = file.read()
        if json_data == "":
            json_data = "{}"
        self._data = json.loads(json_data)

    def dump(self):
        with open(self.file_path, 'w') as file:
            json.dump(self._data, file)



from typing import Optional, Union

from storages.types import StorageType

from storages.sync_storages.pickle_storage import PickleStorage
from storages.sync_storages.json_storage import JSONStorage
from storages.sync_storages.redis_storage import RedisStorage


def load_storage(storage_type: StorageType,
                 file_path: Optional[str] = None,
                 redis_url: Optional[str] = None,
                 redis_data_key: Optional[str] = None):
    storage = {
        StorageType.PICKLE: PickleStorage,
        StorageType.JSON: JSONStorage,
        StorageType.REDIS: RedisStorage,
    }
    if storage_type in [StorageType.PICKLE, StorageType.JSON]:
        return storage[storage_type](file_path)
    elif storage_type == StorageType.REDIS:
        return storage[storage_type](redis_url, redis_data_key)


Storage = Union[
    PickleStorage, JSONStorage, RedisStorage
]

__all__ = (
    "PickleStorage",
    "JSONStorage",
    "RedisStorage",
    "Storage",
    "load_storage",
)



import json
import atexit
from typing import Optional
from storages.sync_storages.base_nosql_storage import BaseNoSQLStorage

redis_installed = True
try:
    from redis import Redis, ConnectionPool
except ModuleNotFoundError:
    redis_installed = False


def create_connection_pool(redis_url):
    if not redis_installed:
        raise ModuleNotFoundError("Redis is not installed. Install it via 'pip install redis'")
    return ConnectionPool.from_url(redis_url)


class RedisStorage(BaseNoSQLStorage):
    def __init__(self, redis_url: Optional[str], data_key: Optional[str]):
        super().__init__()
        self.redis = create_connection_pool(redis_url if redis_url is not None else "redis://localhost/")
        self.data_key = data_key if data_key is not None else "data"
        self.load()

        atexit.register(self.dump)

    def load(self):
        connection = Redis(connection_pool=self.redis)
        result = connection.get(self.data_key)
        connection.close()
        if result is not None:
            self._data = json.loads(result)

    def dump(self):
        connection = Redis(connection_pool=self.redis)
        connection.set(self.data_key, json.dumps(self._data))
        connection.close()



import os
import pickle
import aiofiles
from typing import Optional
from storages.async_storages.base_nosql_storage import BaseNoSQLStorage


def create_dir(file_path):
    dirs, filename = os.path.split(file_path)
    os.makedirs(dirs, exist_ok=True)
    if not os.path.isfile(file_path):
        with open(file_path, 'wb') as file:
            pickle.dump({}, file, protocol=pickle.HIGHEST_PROTOCOL)


class PickleStorage(BaseNoSQLStorage):
    def __init__(self, file_path: Optional[str]):
        super().__init__()
        self.file_path = file_path if file_path is not None else "./.storages/data.pkl"
        create_dir(self.file_path)

    @classmethod
    async def create(cls, file_path: Optional[str]):
        obj = cls(file_path)
        await obj.load()
        return obj

    async def load(self):
        async with aiofiles.open(self.file_path, 'rb') as file:
            pickle_data = await file.read()
        self._data = pickle.loads(pickle_data)

    async def dump(self):
        async with aiofiles.open(self.file_path, 'wb') as file:
            pickle_data = pickle.dumps(self._data, protocol=pickle.HIGHEST_PROTOCOL)
            await file.write(pickle_data)



import abc
from abc import abstractmethod
import copy


class BaseNoSQLStorage(abc.ABC):
    def __init__(self):
        self._data = {}

    @abstractmethod
    async def load(self):
        raise NotImplementedError

    @abstractmethod
    async def dump(self):
        raise NotImplementedError

    async def set_data(self, **kwargs):
        updated = False
        for key, value in kwargs.items():
            if key not in self._data.keys() or self._data[key] != value:
                self._data[key] = copy.deepcopy(value)
                updated = True
        if updated:
            await self.dump()

    async def delete_data(self, *keys):
        for key in keys:
            if key in self._data.keys():
                del self._data[key]
            else:
                raise KeyError(key)
        await self.dump()

    async def reset_data(self):
        self._data = {}
        await self.dump()

    async def get_data(self, *keys) -> tuple:
        if len(keys) == 0:
            return self._data,
        result = []
        for key in keys:
            if isinstance(key, list):
                sub_data = self._data
                for idx, subkey in enumerate(key):
                    if subkey in sub_data.keys():
                        sub_data = sub_data[subkey]
                    else:
                        raise KeyError("->".join(key[:idx + 1]))
                result.append(sub_data)
            elif key in self._data.keys():
                result.append(self._data[key])
            else:
                raise KeyError(key)
        return tuple(result)



import json
from typing import Optional
from storages.async_storages.base_nosql_storage import BaseNoSQLStorage

redis_installed = True
try:
    import aioredis
except ModuleNotFoundError:
    try:
        from redis import asyncio as aioredis
    except ModuleNotFoundError:
        redis_installed = False


def create_connection_pool(redis_url):
    if not redis_installed:
        raise ModuleNotFoundError("AioRedis is not installed. Install it via 'pip install aioredis'")
    return aioredis.ConnectionPool.from_url(redis_url)


class RedisStorage(BaseNoSQLStorage):
    def __init__(self, redis_url: Optional[str], data_key: Optional[str]):
        super().__init__()
        self.redis = create_connection_pool(redis_url if redis_url is not None else "redis://localhost/")
        self.data_key = data_key if data_key is not None else "data"

    @classmethod
    async def create(cls, file_path: Optional[str], data_key: Optional[str]):
        obj = cls(file_path, data_key)
        await obj.load()
        return obj

    async def load(self):
        connection = aioredis.Redis(connection_pool=self.redis)
        result = await connection.get(self.data_key)
        await connection.close()
        if result is not None:
            self._data = json.loads(result)

    async def dump(self):
        connection = aioredis.Redis(connection_pool=self.redis)
        await connection.set(self.data_key, json.dumps(self._data))
        await connection.close()
